import time

from ufotest.testing import AbstractTest, MessageTestResult, AssertionTestResult, TestRunner
from ufotest.camera import pci_read, pci_write
from ufotest.util import clean_pci_read_output


class PowerUpTest(AbstractTest):

    name = 'power_up'

    def __init__(self, test_runner: TestRunner):
        AbstractTest.__init__(self, test_runner)

    def run(self):
        test_result = AssertionTestResult()

        writes = [
            ('9000', 'DFFF'),
            ('9000', 'E0FF'),
            ('9000', 'E10F')
        ]

        for address, value in writes:
            pci_write(address, value)
            read = pci_read('9010', '1')
            cleaned_read = clean_pci_read_output(read)
            test_result.assert_pci_read_ok('write {}, {} --> {}'.format(address, value, cleaned_read))
            time.sleep(0.2)

        return test_result




import smtplib
import ssl
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart

import click

from ufotest.config import Config
from ufotest.util import get_template
from ufotest.testing import TestReport
from ufotest.ci.build import BuildReport

CONFIG = Config()


def send_report_mail(receiver_email: str, receiver_name: str, build_report: BuildReport):
    message = MIMEMultipart('alternative')
    message['Subject'] = 'UfoTest Report'
    message['To'] = receiver_email

    context = {
        'name': receiver_name,
        'report': build_report,
        'config': CONFIG
    }

    text_template = get_template('report_mail.text')
    text = text_template.render(context)
    text_mime = MIMEText(text, 'plain')
    message.attach(text_mime)

    html_template = get_template('report_mail.html')
    html = html_template.render(context)
    html_mime = MIMEText(html, 'html')
    message.attach(html_mime)

    send_mime_multipart(receiver_email, message)
    click.secho('(+) Sent report email to: {}'.format(receiver_email), fg='green')


def send_mime_multipart(receiver_email: str, message: MIMEMultipart):
    sender_email = CONFIG.get_email_address()
    sender_password = CONFIG.get_email_password()

    message['From'] = sender_email

    context = ssl.create_default_context()
    with smtplib.SMTP_SSL('smtp.gmail.com', 465, context=context) as server:
        server.login(sender_email, sender_password)
        server.sendmail(
            sender_email,
            receiver_email,
            message.as_string()
        )



from ufotest.testing import AbstractTest, TestRunner
from ufotest.testing import MessageTestResult


class LibucaInstalledTest(AbstractTest):

    name = 'libuca_installed'

    def __init__(self, test_runner: TestRunner):
        AbstractTest.__init__(self, test_runner)

    def run(self):
        return MessageTestResult(0, 'to be implemented')



import os
import sys
from typing import Optional

import click
import shutil

from ufotest.config import Config, get_path
from ufotest.util import execute_command
from ufotest.install import git_clone, get_repository_name
from ufotest.testing import TestRunner

CONFIG = Config()
UFOTEST_PATH = get_path()


def is_building() -> bool:
    """Returns the boolean value of whether or not a build process for the remote repo is currently running.

    :return: the boolean value of whether a build is currently running or not.
    """
    # So the build process works as follows: The repo folder cloned from the remote location, the new bit file is
    # flashed to the hardware, the test suite is executed and most importantly the afterwards the cloned folder is
    # deleted again! We are using this here to check if a build is currently still running, by checking if this repo
    # folder still exists.
    repository_url = CONFIG.get_ci_repository_url()
    repository_name = get_repository_name(repository_url)
    repository_path = os.path.join(UFOTEST_PATH, repository_name)

    return os.path.exists(repository_path) and os.path.isdir(repository_path)


def build_repository(suite: str, verbose: Optional[bool] = False):
    """Clones the remote repo, flashes the new configuration to the hardware and runs the test *suite*

    :param suite: the string name of the test suite which is supposed to be executed with the camera after the new
    configuration was flashed to the hardware.
    :param verbose: Whether or not additional console output is to be generated.

    :return: void
    """
    repository_path = clone_repository(verbose)

    try:
        # -- FLASHING THE BITFILE TO THE HARDWARE
        # For the flashing process there already exists an CLI command within this very application. So the simplest
        # thing is to just invoke this command to do the flashing process here.
        bitfile_path = os.path.join(repository_path, CONFIG.get_ci_bitfile_path())
        exit_code = execute_command('ufotest flash {}'.format(bitfile_path), True)
        if exit_code:
            # TODO: there seems to be a problem where the flash command does not reurn exit code 1 even with an error.
            click.secho('(+) There was a problem flashing the bitfile!')
            sys.exit(1)
        click.secho('(+) New bitfile flashed to the hardware', fg='green')

        # -- RUNNING THE TEST SUITE
        test_runner = TestRunner()
        test_runner.load()
        click.secho('    Tests have been loaded from memory')

        click.secho('    Running test suite: {}'.format(suite))
        test_report = test_runner.run_suite(suite)
        click.secho('(+) Test report saved to "{}"'.format(test_runner.folder_path), fg='green')

        # -- COPY THE BIT FILE INTO THE ARCHIVE
        bitfile_archive_path = os.path.join(test_runner.folder_path, 'snapshot.bit')
        shutil.copy(bitfile_path, bitfile_archive_path)
        click.secho('(+) Copied the used version of the bitfile: {}'.format(bitfile_archive_path), fg='green')

    except Exception as e:
        click.secho('[!] Exception: {}'.format(str(e)), fg='red')

    finally:
        # At the end it is important that we remove the cloned repo again, so that the next build process can clone the
        # new repo version into the same path.
        shutil.rmtree(repository_path)
        click.secho('(+) deleted repository folder: {}'.format(repository_path), fg='green')


def clone_repository(verbose: Optional[bool] = False):
    """Clones the repository for the camera source code which was specified in the config file.

    The repository is cloned into the base folder of the ufotest installation folder.

    :return: The string of the absolute path for the repository top level folder on the local machine
    """
    # The URL of the git repository is saved in the config file. We will install the repository directly into the folder
    # of the ufotest installation. This will be the same path every time. Due to this fact it is important that we make
    # sure that the installation is deleted at the end of each build process!
    repository_url = CONFIG.get_ci_repository_url()
    branch = CONFIG.get_ci_branch()
    _, repository_path = git_clone(UFOTEST_PATH, repository_url, verbose, branch=branch)

    return repository_path



"""
A module, which contains the code related to the installation process of the UFO dependencies
"""
import os
import re
import click
import subprocess
from typing import Optional, Tuple

from ufotest.config import CONFIG
from ufotest.util import execute_command, get_repository_name


def git_clone(path: str, git_url: str, verbose: bool, branch: str = 'master') -> Tuple[str, str]:
    """Clones the repository from *git_url* into the given folder *path*

    The additional arg *verbose* is a boolean which controls whether the output of this operation is being written to
    stdout stream.

    :param path: The path of the folder into which the repository is supposed to be clonedS
    :param git_url: The string url of the git repository from which to clone
    :param verbose: Whether or not to print additional output to the stdout stream
    :param branch: The string name of the branch to be cloned. Default is master branch.

    :return: A tuple of two values, where the first one is the string name of the repository and the second is the
    string of the absolute path of the clones location within the filesystem.
    """
    # Here we are first extracting the name of the git repository, because that will also be the name of the folder
    # into which it was cloned into later on and this that will be important to actually enter this folder
    repository_name = get_repository_name(git_url)
    repository_path = os.path.join(path, repository_name)
    if verbose:
        click.secho('   Cloning git repository "{}"'.format(git_url))

    # Executing the clone command
    clone_command = 'git clone --single-branch --branch {} {}'.format(branch, git_url)
    exit_code = execute_command(clone_command, verbose, cwd=path)
    if not exit_code:
        click.secho('(+) Cloned repository "{}" ({})'.format(repository_name, repository_path), fg='green')

    return repository_name, repository_path


def install_generic_cmake(path: str, git_url: str, verbose: bool, cmake_args: dict):
    """
    Installs the cmake project from the "git_url" into the given folder "path".

    This function first clones the repository which is given by the git url, then it enters this folder locally,
    creates a build folder and attempts to run a cmake installation process within this folder.
    The cmake_args can be used to pass additional options to the cmake build process.
    """
    # Cloning the repository
    name, folder_path = git_clone(path, git_url, verbose)

    arguments = ' '.join(['-D{}={}'.format(key, value) for key, value in cmake_args.items()])
    build_command = 'mkdir build; cd build; cmake {} ..'.format(arguments)
    exit_code = execute_command(build_command, verbose, cwd=folder_path)
    if not exit_code:
        click.secho('Built "{}" sources'.format(name), fg='green')

    install_command = 'cd build; sudo make install'
    exit_code = execute_command(install_command, verbose, cwd=folder_path)
    if not exit_code:
        click.secho('Installed "{}" successfully!'.format(name), bold=True, fg='green')
    else:
        click.secho('Error installing "{}"!'.format(name), bold=True, fg='red')

    return folder_path


def install_package(package_name: str, verbose=True):
    """
    Installs a system package with the given "package_name"
    """
    operating_system = CONFIG['install']['os']
    package_install_command = CONFIG['install'][operating_system]['package_install']

    click.secho('Installing package "{}"...'.format(package_name))

    command = '{} {}'.format(package_install_command, package_name)
    exit_code = execute_command(command, verbose)

    if not exit_code:
        click.secho('Successfully installed "{}"'.format(package_name), fg='green')
        return True
    else:
        click.secho('Encountered an error while installing "{}"'.format(package_name), fg='yellow')
        return False


def install_dependencies(verbose=True):
    """
    Installs all the system packages, which are listed in the config file.
    """
    operating_system = CONFIG['install']['os']
    packages = CONFIG['install'][operating_system]['packages']

    installed_packages = {}

    for package_name in packages:

        success = install_package(package_name, verbose=verbose)
        installed_packages[package_name] = success

    successful_packages = [package_name for package_name, success in installed_packages.items() if success]
    click.secho('Installed {} packages: {}'.format(
        len(successful_packages),
        ', '.join(successful_packages)
    ), fg='green', bold=True)


def install_fastwriter(path: str, verbose=True):
    """
    Installs the fastwriter repository into the given "path"
    """
    git_url = CONFIG['install']['fastwriter_git']
    install_generic_cmake(
        path,
        git_url,
        verbose,
        {'CMAKE_INSTALL_PREFIX': '/usr'}
    )


def install_pcitools(path:str, verbose=True):
    """
    Installs the pcitool repository into the given "path"
    """
    git_url = CONFIG['install']['pcitools_git']
    folder_path = install_generic_cmake(
        path,
        git_url,
        verbose,
        {'CMAKE_INSTALL_PREFIX': '/usr'}
    )

    # Also installing the driver!
    driver_path = os.path.join(folder_path, 'driver')
    output = None if verbose else subprocess.DEVNULL

    build_command = 'mkdir build; cd build; cmake -DCMAKE_INSTALL_PREFIX=/usr ..'
    exit_code = execute_command(build_command, verbose, cwd=driver_path)
    if not exit_code:
        click.secho('Built "pcilib driver" sources', fg='green')

    install_command = 'cd build; sudo make install'
    exit_code = execute_command(install_command, verbose, cwd=driver_path)
    if not exit_code:
        click.secho('Installed "pcilib driver" successfully!', bold=True, fg='green')

    # Activating the driver after it has been installed...
    activate_driver_command = 'sudo depmod -a'
    exit_code = execute_command(activate_driver_command, verbose)
    if not exit_code:
        click.secho('Activated driver!', bold=True, fg='green')


def install_libufodecode(path:str, verbose=True):
    """
    Installs the libufodecode repository into the given "path"
    """
    git_url = CONFIG['install']['libufodecode_git']
    install_generic_cmake(
        path,
        git_url,
        verbose,
        {
            'CMAKE_INSTALL_PREFIX': '/usr',
            'IPECAMERA_WIDTH': CONFIG.get_sensor_width(),
            'IPECAMERA_HEIGHT': CONFIG.get_sensor_height()
        }
    )


def install_libuca(path: str, verbose=True):
    """
    Installs the libuca repository into the given "path"
    """
    git_url = CONFIG['install']['libuca_git']
    install_generic_cmake(
        path,
        git_url,
        verbose,
        {'CMAKE_INSTALL_PREFIX': '/usr'}
    )


def install_uca_ufo(path: str, verbose=True):
    """
    Installs the uca-ufo repository into the given "path"
    """
    git_url = CONFIG['install']['ucaufo_git']
    install_generic_cmake(
        path,
        git_url,
        verbose,
        {
            'CMAKE_INSTALL_PREFIX': '/usr',
            'CMOSIS_SENSOR_WIDTH': CONFIG.get_sensor_width(),
            'CMOSIS_SENSOR_HEIGHT': CONFIG.get_sensor_height()
        }
    )


def install_ipecamera(path: str, verbose=True):
    """
    Installs the ipecamera repository into the given "path"
    """
    git_url = CONFIG['install']['ipecamera_git']
    install_generic_cmake(
        path,
        git_url,
        verbose,
        {
            'CMAKE_INSTALL_PREFIX': '/usr',
        }
    )


def install_vivado(path: str, verbose=True):
    # Downloading the zip file

    # Unzipping the file into a folder

    # Going into the folder and starting the installation process

    # Returning the path to the folder
    pass



from __future__ import annotations
import os
import click
import datetime
import shutil
import json
from contextlib import AbstractContextManager
from typing import Optional

from ufotest.config import Config, get_path, get_builds_path
from ufotest.util import (AbstractRichOutput,
                          get_repository_name,
                          execute_command,
                          get_command_output,
                          get_template,
                          get_version)
from ufotest.testing import TestRunner, TestReport, TestContext


UFOTEST_PATH = get_path()
BUILDS_PATH = get_builds_path()


# FUNCTIONS

def build_context_from_config(config: Config) -> BuildContext:
    """Creates a new BuildContext instance from the provided *config*.

    This function is mainly for the purpose of a manual build trigger. For manually triggering a build there
    obviously is no request data from which to get the repo url / commit hash from, which is why it will use the
    information which were defined in the config file.

    :return: The build context which was based on the information in the config file in the ufotest folder
    """
    repository_url = config.get_ci_repository_url()
    repository_branch = config.get_ci_branch()
    repository_commit = 'FETCH_HEAD'
    test_suite = config.get_ci_suite()

    return BuildContext(
        repository_url,
        repository_branch,
        repository_commit,
        test_suite
    )


def build_context_from_request(data: dict) -> BuildContext:
    """Creates a new BuildContext instance from given json payload *data* from the github web hook request.

    This function extracts the information about which commit to clone from which repo from the given payload of the
    request.

    :return: the build context instance.
    """
    repository_url = data['repository']['clone_url']
    repository_branch = data['ref'].replace('refs/heads/', '')
    repository_commit = data['commits'][-1]

    config = Config()
    test_suite = config.get_ci_suite()

    return BuildContext(
        repository_url,
        repository_branch,
        repository_commit,
        test_suite
    )


# CLASSES

class BuildLock(object):
    """This is a static class which models a persistent lock, which keeps multiple build processes from running at the
    same time.

    **Why?**
    So first of all why is this important. The ci server of ufotest accepts request which are generated by github
    web hooks, whenever new code is pushed to source repository. As a response to such a push, a new build process is
    initiated. So assuming the case that there would be two separate pushes to the repo in quick succession. This
    would cause a second python process to attempt to start a build process before the first one is done. The build
    process interfaces with the actual camera hardware. Having two processes attempt to send commands to the same
    hardware would result in errors. Thus, the ufotest app has to be restricted to only one build process at a time.

    **How?**
    This static class implements this locking functionality by providing the 'acquire' and 'release' methods. With
    these a indicator file 'build.locked' is saved to the installation folder of the ufotest installation. As long as
    this file exists, this locks status will be blocking. The release method simply removes this file again.
    """
    LOCK_PATH = os.path.join(UFOTEST_PATH, 'build.locked')

    @classmethod
    def acquire(cls):
        """Acquires the lock for the process which calls this method.

        If the lock is already locked by another process, this method will raise an error.

        :raise PermissionError: If another build process has already acquired the lock.
        """
        if cls.is_locked():
            raise PermissionError('Another build process is currently already running')
        else:
            with open(cls.LOCK_PATH, mode='w') as lock_file:
                lock_file.write('True')

    @classmethod
    def release(cls):
        """Releases the locK again, so that other processes can start build processes in the future.

        :return: void
        """
        os.remove(cls.LOCK_PATH)

    @classmethod
    def is_locked(cls) -> bool:
        """Returns the boolean value of whether or not the lock is currently locked by some process.

        If the lock is indeed locked, this would mean that currently there is a build process running.

        :return: a boolean
        """
        return os.path.exists(cls.LOCK_PATH)


class BuildContext(AbstractContextManager):
    """Represents the context of a build process. The context in this case refers to all necessary information about
    and around the build process. Objects of this context class will save all the information about the build process
    and manage all the secondary processes which are involved with a build process such as the acquisition of the lock
    and the creation of the build archive folder. But the actual build process is NOT the responsibility of this class!

    **Context Manager**

    This class is a python context manager, which means that it implements the __enter__ and __exit__ magic methods and
    can be used with the 'with' keyword.
    The __enter__ of the context acquires the build lock for the current process and creates the correct folder within
    the 'builds' archive.
    The __exit__ of the context releases the lock again and deletes the folder of the actual cloned repo.
    It is important that both these functions actually get called, which is why this class should absolutely be used
    in combination with the 'with' keyword!

    .. code-block:: python

        with BuildContext(*args) as context:
            # only here the context should actually be used!

    """
    def __init__(self, repository_url: str, branch_name: str, commit_name: str, test_suite: str):
        # constructed attributes
        self.repository_url = repository_url
        self.branch = branch_name
        self.commit = commit_name
        self.test_suite = test_suite

        # calculated attributes
        self.config = Config()
        self.creation_datetime = datetime.datetime.now()
        self.version = get_version()
        self.test_context = TestContext()

        # derived attributes
        self.repository_name = get_repository_name(repository_url)
        self.repository_path = os.path.join(UFOTEST_PATH, self.repository_name)
        self.folder_name = '{}__{}'.format(
            self.repository_name,
            self.creation_datetime.strftime('%Y_%m_%d__%H_%M')
        )
        self.folder_path = os.path.join(BUILDS_PATH, self.folder_name)

        # non initialized attributes
        # These are the attributes, which have to be declared, but which cannot be initialized in the constructor.
        # The start and end time of course can only be acquired during the actual start and end.
        # The bitfile path actually has to be set by the BuildRunner. It is not known in before hand! The test report
        # as well can obviously only be added after the actual test are done.
        self.start_datetime: Optional[datetime.datetime] = None
        self.end_datetime: Optional[datetime.datetime] = None
        self.bitfile_path = None
        self.test_report: Optional[TestReport] = None

    # IMPLEMENT 'AbstractContextManger'

    def __enter__(self) -> BuildContext:
        """Magic method which gets called whenever the context manager is entered. Returns the context manager instance.

        This method gets called whenever the context is entered. It manages two important responsibilities in
        preparation for the build process:

        - It acquires the build lock for the current process.
        - It creates the according folder for the build process within the 'builds' archive.

        :raise PermissionError: If another build process is currently running.

        :return: The object instance of BuildContext on which this method was invoked
        """
        # -- ACQUIRING THE LOCK
        # So the BuildLock is a static class which models a *persistent* lock which is supposed to prevent multiple
        # parallel build processes to be running at the same time. Since there is only one hardware every process would
        # try to interface with, this would be a bad idea. I also contemplated whether I should already acquire the
        # lock here or during the actual build process. The final knockout argument was that since i am creating the
        # folder here, if the lock acquisition failed later on, the folder would still be created and that would not be
        # the behaviour we want.
        BuildLock.acquire()

        # -- CREATING THE BUILD FOLDER
        os.mkdir(self.folder_path)

        # -- RECORDING THE START TIME
        self.start_datetime = datetime.datetime.now()

        # -- ENTERING TEST CONTEXT
        self.test_context = self.test_context.__enter__()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Magic method which gets called whenever the context is left again.

        This method gets called whenever the context is being closed. It manages two important responsibilities for
        finalizing the build process:

        - It releases the build lock
        - It deletes the folder of the cloned repository, so that the next process can perform an entirely new clone
          process.

        :return: void
        """
        # -- DELETING THE CLONED REPOSITORY
        # another important thing to do to be able to have another build process be able to run again is to delete the
        # folder which was used for the cloning process.
        shutil.rmtree(self.repository_path)

        # -- RELEASING THE LOCK
        # By doing this, another build process is now able to be started again in a different process possibly
        BuildLock.release()

        # -- EXIT TEST CONTEXT
        self.test_context.__exit__(exc_type, exc_val, exc_tb)


class BuildReport(AbstractRichOutput):
    """This class represents a report, which is being created as the result of a build process. A build report contains
    information and metrics about the build process which are important to provide to the users.

    *Rich Output*

    This class implements the abstract base class AbstractRichOutput, which means that it can be converted to various
    formats, which include markdown and html. These can be saved to files to later be reviewed by a user.
    """
    DATETIME_FORMAT = '%d.%m.%Y %H:%M'

    def __init__(self, build_context: BuildContext):
        # constructed attributes
        self.context = build_context

        # derived attributes from context
        self.start = self.context.start_datetime.strftime(self.DATETIME_FORMAT)
        self.end = self.context.end_datetime.strftime(self.DATETIME_FORMAT)
        duration_time_delta = self.context.end_datetime - self.context.start_datetime
        self.duration = round(duration_time_delta.seconds / 60)
        self.repository = self.context.repository_url
        self.repository_name = self.context.repository_name
        self.branch = self.context.branch
        self.commit = self.context.commit
        self.folder = self.context.folder_path
        self.folder_name = os.path.basename(self.folder)
        self.bitfile_path = self.context.bitfile_path
        self.bitfile_name = os.path.basename(self.bitfile_path)
        self.test_suite = self.context.test_suite
        self.test_count = self.context.test_report.test_count
        self.test_success_count = self.context.test_report.successful_count
        self.test_percentage = self.context.test_report.success_ratio
        self.test_folder_name = self.context.test_report.folder_name

    def save(self, folder_path: str):
        """Saves the both the MD and HTML version of the report into the given *folder_path*.

        :return: void
        """
        # 1 -- SAVE MARKDOWN FILE
        markdown_path = os.path.join(folder_path, 'report.md')
        with open(markdown_path, mode='w') as markdown_file:
            markdown_file.write(self.to_markdown())

        # 2 -- SAVE HTML FILE
        html_path = os.path.join(folder_path, 'report.html')
        with open(html_path, mode='w') as html_file:
            html_file.write(self.to_html())

        # 3 -- SAVE JSON FILE
        json_path = os.path.join(folder_path, 'report.json')
        with open(json_path, mode='w') as json_file:
            json_file.write(self.to_json())

        click.secho('(+) Build report saved to: {}'.format(folder_path), fg='green')

    def __str__(self):
        return self.to_string()

    # == JSON SAVING

    def to_dict(self) -> dict:
        return {
            'start':                self.start,
            'end':                  self.end,
            'duration':             self.duration,
            'version':              self.context.version,
            'test_suite':           self.test_suite,
            'test_count':           self.test_count,
            'test_success_count':   self.test_success_count,
            'test_percentage':      self.test_percentage,
            'test_folder_name':     self.test_folder_name,
            'repository':           self.repository,
            'repository_name':      self.repository_name,
            'branch':               self.branch,
            'commit':               self.commit,
            'folder_name':          self.folder_name,
            'bitfile_path':         self.bitfile_path,
        }

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    # == AbstractRichOutput

    def to_markdown(self) -> str:
        template = get_template('build_report.md')
        return template.render({'report': self})

    def to_latex(self) -> str:
        # not important right now
        pass

    def to_string(self) -> str:
        template = get_template('build_report.text')
        return template.render({'report': self})

    def to_html(self) -> str:
        template = get_template('build_report.html')
        return template.render({'report': self})


class BuildRunner(object):
    """This class is responsible for actually executing the build process.

    This build process involves the following individual steps:

    - The remote git repository is cloned to the local system and checked out to the desired branch/commit.
    - The actual bit file is extracted from this repo.
    - This bit file is being used to flash the new configuration to the fpga board.
    - Finally the desired test suite is being run with this new hardware version.

    A new BuildRunner object needs to be constructed with an appropriate BuildContext object as an argument. This is
    because the context object contains all the necessary information about *how* the build process is supposed to work
    in terms of which repo to clone, which test suite to run etc.

    **Example**

    A typical build process would look something like this:

    .. code-block:: python

        with BuildContext(*args) as build_context:
            build_runner = BuildRunner(build_context)
            build_report: BuildReport = build_runner.build()
            build_report.save()

    """
    def __init__(self, build_context: BuildContext):
        self.context = build_context
        self.test_context = self.context.test_context

        self.config = Config()
        self.test_runner = TestRunner(self.test_context)

    def build(self) -> BuildReport:
        """Actually performs the build process and returns a BuildReport object about the result.

        This build process involves the following individual steps:

        - The remote git repository is cloned to the local system and checked out to the desired branch/commit.
        - The actual bit file is extracted from this repo.
        - This bit file is being used to flash the new configuration to the fpga board.
        - Finally the desired test suite is being run with this new hardware version.

        :return: A BuildReport object which contains all the informations and metrics about the build process which are
        important for the user.
        """
        try:
            click.secho('| | BUILDING FROM SOURCE | |', bold=True)
            # -- CLONING THE REPO
            self.clone()

            # -- FLASHING THE BIT FILE
            self.copy_bitfile()
            self.flash()

            # -- RUNNING THE TEST SUITE
            self.test()

            # -- SETTING THE END TIME
            self.context.end_datetime = datetime.datetime.now()

        except Exception as e:
            click.secho('[!] Build Exception: {}'.format(str(e)), fg='red')
        finally:
            return BuildReport(self.context)

    def test(self):
        """Executes the test suite with the new hardware version

        :return: void
        """
        # -- RUN THE TESTS
        self.test_runner.load()
        click.secho('    Tests have been loaded from memory')

        self.test_runner.run_suite(self.context.test_suite)
        test_report = TestReport(self.test_context)
        click.secho('    Executed test suite: {}'.format(self.context.test_suite))
        click.secho('(+) Test report saved to: {}'.format(self.test_context.folder_path), fg='green')

        # -- STORE THE TEST REPORT
        # The test report later also needs to be referenced to produce the build report, so it needs to be saved as
        # part of the build context object as well. The context object already has the attribute 'test_report', which
        # was initialized to none in the constructor.
        self.context.test_report = test_report

    def flash(self):
        """Flashes the bit file to the fpga board.

        NOTE: For this method to work properly, the attribute 'bitfile_path' of the context object has to be set
        correctly already. This is not being done in this method

        :return: void
        """
        # For the flashing process there already exists an CLI command within this very application. So the simplest
        # thing is to just invoke this command to do the flashing process here.
        exit_code = execute_command('ufotest flash {}'.format(self.context.bitfile_path), True)
        if exit_code:
            # TODO: there seems to be a problem where the flash command does not reurn exit code 1 even with an error.
            click.secho('[!] There was a problem flashing the bitfile!', fg='red')
        click.secho('(+) New bitfile flashed to the hardware', fg='green')

    def copy_bitfile(self):
        """Copies the bitfile from the cloned repo to the build folder for this build.

        The responsibility of this function is to find the bit file (this is the file which can be used to actually
        flash the new configuration onto the hardware) within the cloned repo folder. This bitfile then has to be
        copied from the repo to the build folder so it is stored for later reference. Also the path to this bit file
        has to be stored as an attribute of the context object.

        :return: void
        """
        # -- FIND THE BITFILE
        # So before we can copy the bitfile we actually have to know where is is within the repo folder. This is not
        # such a trivial task. It would kind of be desirable if it were possible to discover it automatically in the
        # future. The way it currently works is that the relative path to this bit file has to be given in the ufotest
        # config file and this relative path will always be used to *hopefully* point at the correct bitfile.
        relative_path = self.config.get_ci_bitfile_path()
        bitfile_path = os.path.join(self.context.repository_path, relative_path)
        bitfile_name = os.path.basename(bitfile_path)

        # -- COPY IT INTO THE FOLDER
        bitfile_destination_path = os.path.join(self.context.folder_path, bitfile_name)
        shutil.copy(bitfile_path, bitfile_destination_path)

        # -- SET CONTEXT ATTRIBUTE
        # The context object actually already has the attribute 'bitfile_path'. It has been initialized as None within
        # the constructor.
        self.context.bitfile_path = bitfile_destination_path

    def clone(self):
        """Clones the remote git repository to the local machine and checking out the desired commit.

        :return: void
        """
        # TODO: Maybe raise some exceptions in this method.
        click.secho('    Cloning git repository: {}'.format(self.context.repository_url))

        # -- CLONING THE REPO
        clone_command = 'git clone --single-branch --branch "{}" "{}"'.format(
            self.context.branch,
            self.context.repository_url
        )
        exit_code = execute_command(clone_command, False, cwd=UFOTEST_PATH)
        click.secho('(+) Cloned git repo: {}'.format(self.context.repository_path), fg='green')

        # -- CHECKOUT CORRECT COMMIT
        checkout_command = 'git checkout "{}"'.format(self.context.commit)
        exit_code = execute_command(checkout_command, False, cwd=self.context.repository_path)
        click.secho('(+) Checking out commit: {}'.format(self.context.commit))

        # -- GET THE COMMIT NAME
        # Ok so here we are doing something weird... Quick explanation: The build process can be triggered without
        # an actual commit name. In that case, it will just use the most recent commit by specifying FETCH_HEAD. This
        # would actually be the commit name within the context object. And that works fine and stuff, but for the
        # report it would be nice if we had the actual commit name. This we can only actually get after we have
        # cloned the repo. So this command here gets the commit name and returns it. We then update the commit name
        # within the context object with this actual name, so that we can use this later in the report.
        hash_command = 'git rev-parse HEAD'
        commit_name = get_command_output(hash_command, cwd=self.context.repository_path).rstrip('\n')
        self.context.commit = commit_name





from ufotest.testing import AbstractTest, TestRunner, MessageTestResult


class MockTest(AbstractTest):

    name = "mock"

    description = (
        "This test case is simply a mock, which is used to test the ufotest software tools. The test case does "
        "absolutely nothing and returns a simple MessageTestResult."
    )

    def __init__(self, test_runner: TestRunner):
        AbstractTest.__init__(self, test_runner)

    def run(self):
        return MessageTestResult(0, 'Mock returned 0')



import os
import json
import subprocess
from multiprocessing import Process

from flask import Flask, request, send_from_directory, jsonify

from ufotest.config import Config, get_path
from ufotest.util import get_template
from ufotest.ci.build import BuildLock, BuildRunner, build_context_from_request
from ufotest.ci.mail import send_report_mail

CONFIG = Config()
PATH = get_path()

ARCHIVE_PATH = os.path.join(PATH, 'archive')
BUILDS_PATH = os.path.join(PATH, 'builds')
STATIC_PATH = os.path.join(PATH, 'static')

server = Flask('UfoTest CI Server', static_folder=None)


@server.route('/', methods=['GET'])
def home():
    template = get_template('home.html')
    return template.render({}), 200


@server.route('/push/github', methods=['POST'])
def push():
    data = request.get_json()

    if BuildLock.is_locked():
        # error code 423 represents the fact that the resource which is being requested is locked. I think this best
        # represents what is going on. There cannot be two simultaneous processes because both of them would be trying
        # to access the same hardware. This is why a request will be rejected if another build process is
        # currently running.
        return 'A build process is currently in progress!', 423

    def build(_data):
        # -- ACTUALLY BUILD
        with build_context_from_request(_data) as build_context:
            build_runner = BuildRunner(build_context)
            build_report = build_runner.build()
            build_report.save(build_context.folder_path)

        # -- SEND REPORT MAILS
        # After the build process has terminated we want to inform the relevant people of the outcome. This is mainly
        # Two people: The maintainer of the repository is getting an email and the person which actually issued the
        # push is getting an email.
        maintainer_email = _data['repository']['owner']['email']
        maintainer_name = _data['repository']['owner']['name']
        send_report_mail(maintainer_email, maintainer_name, build_report)

        pusher_email = _data['pusher']['email']
        pusher_name = _data['pusher']['name']
        send_report_mail(pusher_email, pusher_name, build_report)

    process = Process(target=build, args=(data, ))
    process.start()

    return 'New build process was initiated', 200


@server.route('/archive')
def archive_list():
    reports = []
    for root, folders, files in os.walk(ARCHIVE_PATH):
        for folder in folders:
            folder_path = os.path.join(root, folder)
            report_json_path = os.path.join(folder_path, 'report.json')
            with open(report_json_path, mode='r') as report_json_file:
                report = json.loads(report_json_file.read())
                reports.append(report)

        break

    template = get_template('archive_list.html')
    return template.render({'reports': reports}), 200


@server.route('/archive/<path:path>')
def archive_detail(path):
    return send_from_directory(ARCHIVE_PATH, path)


@server.route('/builds')
def builds_list():
    reports = []
    for root, folders, files in os.walk(BUILDS_PATH):
        for folder in folders:
            folder_path = os.path.join(root, folder)
            report_json_path = os.path.join(folder_path, 'report.json')
            with open(report_json_path, mode='r') as report_json_file:
                report = json.loads(report_json_file.read())
                reports.append(report)

        break

    template = get_template('builds_list.html')
    return template.render({'reports': reports}), 200


@server.route('/builds/<path:path>')
def builds_detail(path):
    return send_from_directory(BUILDS_PATH, path)


@server.route('/static/<path:path>')
def static(path):
    return send_from_directory(STATIC_PATH, path)



"""
Module containing all the actual console scripts of the project.
"""
import sys
import os
from pathlib import Path

import click
import matplotlib
import numpy as np
import shutil
import matplotlib.pyplot as plt
# matplotlib.use('TkAgg')

from ufotest.config import PATH, get_config_path, Config
from ufotest.scripts import SCRIPTS, SCRIPTS_PATH
from ufotest.util import (execute_command,
                          setup_environment,
                          init_install,
                          check_install,
                          execute_script,
                          get_version,
                          get_path,
                          check_vivado,
                          check_path)
from ufotest.install import (install_dependencies,
                             install_fastwriter,
                             install_pcitools,
                             install_libufodecode,
                             install_libuca,
                             install_uca_ufo,
                             install_ipecamera)
from ufotest.camera import save_frame, import_raw, set_up_camera, tear_down_camera
from ufotest.testing import TestRunner, TestContext, TestReport
from ufotest.ci.build import BuildRunner, BuildContext, BuildReport, build_context_from_config
from ufotest.ci.server import server


CONFIG = Config()


@click.group(invoke_without_command=True)
@click.option('--version', '-v', is_flag=True, help='Print the version of the program')
def cli(version):
    if version:
        version = get_version()
        click.secho('UFOTEST VERSION')
        click.secho(version, bold=True)
        return 0


@click.command('install', short_help='Install the project and its dependencies')
@click.argument('path', type=click.Path(exists=True, file_okay=False, dir_okay=True, writable=True))
@click.option('--verbose', '-v', is_flag=True, help='Show additional console output')
@click.option('--no-dependencies', '-d', is_flag=True, help='Skip installation of required system packages')
@click.option('--no-libuca', '-l', is_flag=True, help='Skip installation of libuca')
@click.option('--no-vivado', is_flag=True, help='Skip installation of vivado')
def install(path, verbose, no_dependencies, no_libuca, no_vivado):
    """
    Installing the Project into PATH

    PATH will be the system path, which will then contain subfolders with all the required repositories and
    dependencies
    """
    if not check_install():
        return 1

    path = os.path.realpath(path)

    operating_system = CONFIG['install']['os']
    click.secho('\n| | STARTING INSTALLATION | |', bold=True)
    click.secho('--| Configured OS: {}'.format(operating_system))
    click.secho('--| Configured package install: {}'.format(CONFIG['install'][operating_system]['package_install']))
    click.secho('--| Camera dimensions: {} x {}'.format(
        CONFIG.get_sensor_width(),
        CONFIG.get_sensor_height()
    ))

    if not no_dependencies:
        click.secho('\n=====| Installing System Packages |=====', bold=True)
        install_dependencies(verbose=verbose)
    else:
        click.secho('\n=====| Skipping Dependencies |=====', bold=True)

    click.secho('\n=====| Installing fastwriter |=====', bold=True)
    install_fastwriter(path, verbose=verbose)

    click.secho('\n=====| Installing pcitools |=====', bold=True)
    install_pcitools(path, verbose=verbose)

    click.secho('\n=====| Installing libufodecode |=====', bold=True)
    install_libufodecode(path, verbose=verbose)

    if not no_libuca:
        click.secho('\n=====| Installing Libuca |=====', bold=True)
        install_libuca(path, verbose=verbose)

        click.secho('\n=====| Installing uca-ufo |=====', bold=True)
        install_uca_ufo(path, verbose=verbose)

        click.secho('\n=====| Installing ipecamera plugin |=====', bold=True)
        install_ipecamera(path, verbose)
    else:
        click.secho('\n=====| Skipping Libuca |=====', bold=True)
    return 0


@click.command('init', short_help='Initializes the installation folder and config file for the app')
@click.option('--verbose', '-v', is_flag=True, help='print additional console messages')
@click.option('--force', '-f', is_flag=True, help='Deletes the current installation to reinstall')
def init(verbose, force):
    """Initializes the installation folder for this application. This folder will be located at *$HOME/.ufotest*.
    This init includes the creation of the necessary folder structure for the archive and the tests, as well as the
    creation of the config file from a default template.
    """
    installation_path = get_path()
    click.secho('\n| | INITIALIZING UFOTEST INSTALLATION | |', bold=True)
    click.secho('--| installation path: {}'.format(installation_path))

    if check_path(installation_path, is_dir=True):
        if force:
            shutil.rmtree(get_path())
            click.secho('(+) Deleted old installation folder!', fg='green')
        else:
            click.secho('[!] An installation folder already exists!', fg='red')
            click.secho('    Please use the --force flag if you wish to replace the existing installation')
            sys.exit(1)
    else:
        click.secho('    Installation folder does not yet exist')

    init_install(verbose=verbose)
    click.secho('(+) UfoTest app is initialized!', bold=True, fg='green')

    sys.exit(0)


@click.command('config', short_help='Edit the config for ufotest')
@click.option('--editor', '-e', type=click.STRING, help='Specify the editor command to be used to open the config file')
def config(editor):
    """
    Edit the configuration file for this project
    """
    if not check_install():
        return 1

    config_path = get_config_path()
    click.edit(filename=config_path, editor=editor)

    return 0


@click.command('frame', short_help='Acquire and display a frame from the camera')
@click.option('--verbose', '-v', is_flag=True, help='print additional console messages')
@click.option('--output', '-o', type=click.STRING,
              help='Specify the output file path for the frame', default='/tmp/frame.raw')
@click.option('--display', '-d', is_flag=True, help='display the frame in seperate window')
def frame(verbose, output, display):
    """
    Capture a frame from the camera and display it to the user
    """
    if not check_install():
        return 1

    # Setup all the important environment variables and stuff
    setup_environment()

    execute_command('rm /tmp/frame*', verbose)
    if verbose:
        click.secho('Removed the previous frame buffer', fg='green')

    # Call the necessary pci commands
    save_frame(output, verbose=verbose)

    # Display the file using matplot lib
    if display:
        images = import_raw(
            path=output,
            n=1,
            sensor_height=CONFIG.get_sensor_height(),
            sensor_width=CONFIG.get_sensor_width()
        )

        plt.imshow(images[0])
        plt.show()

    return 0


@click.command('script', short_help="Execute one of the known (bash) scripts")
@click.argument('name', type=click.STRING)
@click.option('--verbose', '-v', is_flag=True, help='print additional console messages')
def script(name, verbose):
    """Executes a registered script with the given NAME
    """
    if not check_install():
        return 1

    exit_code = execute_script(name, verbose=verbose)
    if not exit_code:
        click.secho('Script "{}" succeeded'.format(name), bold=True, fg='green')
    else:
        click.secho('Script "{}" failed'.format(name), bold=True, fg='red')


@click.command('list-scripts', short_help='List all available scripts')
def list_scripts():
    if not check_install():
        return 1

    click.secho('The following scripts are available:\n', fg='green', bold=True)

    for script_id, script_data in SCRIPTS.items():
        click.secho(script_id, bold=True)
        click.secho('   Path: {}\n   Description: {}\n   Author: {}\n'.format(
            script_data['path'],
            script_data['description'],
            script_data['author']
        ))


@click.command('setup', short_help="Enable the camera")
@click.option('--verbose', '-v', is_flag=True, help='print additional console messages')
def setup(verbose):
    if not check_install():
        return 1

    set_up_camera(verbose=verbose)


@click.command('teardown', short_help="Disable the camera. DO NOT USE")
@click.option('--verbose', '-v', is_flag=True, help='print additional console messages')
def teardown(verbose):
    if not check_install():
        return 1

    tear_down_camera(verbose)


@click.command('flash', short_help='Flash a new .BIT file to the FPGA memory')
@click.option('--verbose', '-v', is_flag=True, help='print additional console messages')
@click.argument('file', type=click.STRING)
def flash(verbose, file: str):
    """
    Flashes a given ".bit" FILE to the internal memory of the fpga

    FILE will be the string path of the file to be used for the flashing process
    """
    if not check_install():
        return 1

    # -- ECHO CONFIGURATION
    click.secho('\n| | FLASHING THE FPGA MEMORY | |', bold=True)
    click.secho('--| bitfile: {}\n'.format(file))

    # -- CHECKING IF THE GIVEN FILE EVEN EXISTS
    file_path = os.path.abspath(file)
    file_exists = check_path(file_path, is_dir=False)
    if not file_exists:
        click.secho('[!] The given path does not exist!', fg='red')
        return 1

    # -- CHECKING IF THE FILE EVEN IS A BIT FILE
    file_extension = file_path.split('.')[-1].lower()
    is_bit_file = file_extension == 'bit'
    if not is_bit_file:
        click.secho('[!] The given file is not a .BIT file', fg='red')
        return 1

    # -- CHECKING IF VIVADO IS INSTALLED
    vivado_installed = check_vivado()
    if not vivado_installed:
        click.secho('[!] Vivado is not installed, please install Vivado to be able to flash the fpga!', fg='red')
        return 1
    elif verbose:
        click.secho('    Vivado installation found')

    # -- STARTING VIVADO SETTINGS
    vivado_command = CONFIG['install']['vivado_settings']
    execute_command(vivado_command, verbose, cwd=os.getcwd())
    click.secho('    Finished setup of vivado environment')

    # -- FLASHING THE BIT FILE
    flash_command = "{command} -nolog -nojournal -mode batch -source fpga_conf_bitprog.tcl -tclargs {file}".format(
        command=CONFIG['install']['vivado_command'],
        file=file_path
    )
    exit_code = execute_command(flash_command, verbose, cwd=SCRIPTS_PATH)
    if not exit_code:
        click.secho('(+) Flashed FPGA with: {}'.format(file_path), fg='green', bold=True)
        return 0
    else:
        return 1


@click.command('test', short_help="Run a camera test")
@click.option('--verbose', '-v', is_flag=True, help='print additional console messages')
@click.option('--suite', '-s', is_flag=True, help='Execute a test SUITE with the given name')
@click.argument('test_id', type=click.STRING)
def test(verbose, suite, test_id):
    """
    Run the test "TEST_ID"

    TEST_ID is a string, which identifies a certain test procedure. To view all these possible identifiers consult the
    config file.

    This command will execute either a single test case for the camera or a test suite, combines multiple test cases.
    The results of these tests will generate a test report. This test report will be saved into the archive of all
    test reports as a markdown and html file.
    """
    if not check_install():
        return 1

    try:
        with TestContext() as test_context:
            # 1 -- DYNAMICALLY LOADING THE TESTS
            test_runner = TestRunner(test_context)
            test_runner.load()

            # 2 -- EXECUTING THE TEST CASES
            if suite:
                click.secho('    Executing test suite: {}...'.format(test_id), bold=True)
                test_runner.run_suite(test_id)
            else:
                click.secho('    Executing test: {}...'.format(test_id), bold=True)
                test_runner.run_test(test_id)

            # 3 -- SAVING THE RESULTS AS A REPORT
            test_report = TestReport(test_context)
            test_report.save(test_context.folder_path)

    except Exception as e:
        click.secho('[!] {}'.format(e), fg='red', bold=True)
        return 1

    if verbose:
        click.secho(test_report.to_string())

    click.secho('(+) Test report saved to: {}'.format(test_context.folder_path), fg='green', bold=True)
    click.secho('    View the report at: http://localhost/archive/{}/report.html'.format(test_context.folder_name))

    return 0


@click.group('ci', short_help='continuous integration related commands')
def ci():
    pass


@click.command('build', short_help='Applies newest commit from remote repository and then executes test suite')
@click.option('--verbose', '-v', is_flag=True, help='print additional console messages')
@click.argument('suite', type=click.STRING)
def build(verbose, suite):
    """
    Start a new CI build process using the test suite SUITE.
    """
    # -- ECHO CONFIGURATION
    click.secho('\n| | INTEGRATING REMOTE REPOSITORY | |', bold=True)
    click.secho('--| repository url: {}'.format(CONFIG.get_ci_repository_url()))
    click.secho('--| repository branch: {}'.format(CONFIG.get_ci_branch()))
    click.secho('--| bitfile: {}\n'.format(CONFIG.get_ci_bitfile_path()))

    # -- RUNNING THE PROCESS
    with build_context_from_config(CONFIG) as build_context:
        build_runner = BuildRunner(build_context)
        build_report = build_runner.build()
        build_report.save(build_context.folder_path)


@click.command('serve', short_help='Runs the CI server which responds to remote repository webhooks')
@click.option('--verbose', '-v', is_flag=True, help='print additional console messages')
@click.argument('hostname', type=click.STRING)
@click.argument('port', type=click.INT)
def serve(verbose, hostname, port):
    """
    Start the CI server. The server listens for incoming build requests and serves the static report files.
    """
    # -- ECHO CONFIGURATION
    click.secho('\n| | STARTING CI SERVER | |', bold=True)
    click.secho('--| repository url: {}'.format(CONFIG.get_ci_repository_url()))
    click.secho('--| repository branch: {}'.format(CONFIG.get_ci_branch()))
    click.secho('--| bitfile: {}\n'.format(CONFIG.get_ci_bitfile_path()))

    # -- MODIFYING CONTEXT
    # The port value will is stored here in the global config object, because it is needed in some subroutine of the
    # serve process.
    CONFIG['context']['hostname'] = hostname
    CONFIG['context']['port'] = port

    click.secho('(+) Visit the server at http://{}:{}/'.format(hostname, port), fg='green')

    # -- STARTING THE SERVER
    server.run(port=port)


ci.add_command(build)
ci.add_command(serve)

# Registering the commands within the click group
cli.add_command(init)
cli.add_command(config)
cli.add_command(install)
cli.add_command(script)
cli.add_command(frame)
cli.add_command(setup)
cli.add_command(teardown)
cli.add_command(list_scripts)
cli.add_command(flash)
cli.add_command(test)
cli.add_command(ci)


if __name__ == "__main__":
    sys.exit(cli())  # pragma: no cover



import inspect

from ufotest._testing import UfotestTestCase
from ufotest.util import random_string
from ufotest.testing import TestRunner, AbstractTest, TestReport, MessageTestResult, TestMetadata, AssertionTestResult


# HELPER FUNCTIONS
# ================

def get_message_test_result(exit_code: int = 0):
    return MessageTestResult(
        exit_code=exit_code,
        message=random_string(20)
    )


# TESTCASES
# =========

class TestTestRunner(UfotestTestCase):

    def test_construction(self):
        """
        Simply tests if the construction of a new instance does not produce errors
        """
        test_runner = TestRunner(config=self.config)
        self.assertIsInstance(test_runner, TestRunner)

    def test_module_loading_working_at_all(self):
        """
        Tests if the dynamic loading of the modules from the test folder works at all. That does not imply that it
        tests for the correct amount of tests.
        """
        test_runner = TestRunner(config=self.config)
        test_runner.load()

        self.assertNotEqual(0, len(test_runner.modules))
        for module in test_runner.modules.values():
            self.assertTrue(inspect.ismodule(module))

    def test_test_loading_working_at_all(self):
        """
        Tests if any tests have been loaded from the test modules and also tests if these are only valid subclasses
        of the AbstractTest base class.
        """
        test_runner = TestRunner(config=self.config)
        test_runner.load()

        self.assertNotEqual(0, len(test_runner.tests))
        for test in test_runner.tests.values():
            self.assertTrue(inspect.isclass(test))
            self.assertTrue(issubclass(test, AbstractTest))

    def test_run_mock_test(self):
        """
        Tests if the "mock" test is being run correctly
        """
        test_runner = TestRunner(config=self.config)
        test_runner.load()

        test_results = test_runner.run_test('mock')
        self.assertIsInstance(test_results, TestReport)
        self.assertEqual(1, test_results.test_count)
        self.assertEqual(1, test_results.passing_count)
        self.assertEqual(0, test_results.error_count)

    def test_run_mock_suite(self):
        """
        Tests if the "mock" test suite is being run correctly
        """
        test_runner = TestRunner(config=self.config)
        test_runner.load()

        test_results = test_runner.run_suite('mock')
        self.assertIsInstance(test_results, TestReport)
        self.assertEqual(1, test_results.test_count)
        self.assertEqual(1, test_results.passing_count)
        self.assertEqual(0, test_results.error_count)


class TestTestReport(UfotestTestCase):

    def test_construction(self):
        """
        If an instance can be constructed without errors
        """
        test_report = TestReport({'test': get_message_test_result()}, TestMetadata())
        self.assertIsInstance(test_report, TestReport)

    def test_markdown_conversion_rough(self):
        """
        If the markdown conversion works at all
        """
        results = {}
        for i in range(5):
            results['test {}'.format(i)] = get_message_test_result()

        test_report = TestReport(results, TestMetadata())
        markdown_string = test_report.to_markdown()

        self.assertIn('# Test Report', markdown_string)


class TestAssertionTestResult(UfotestTestCase):

    def test_construction(self):
        """
        If the construction of an instance generally works
        """
        test_result = AssertionTestResult()
        self.assertIsInstance(test_result, AssertionTestResult)

    def test_adding_assertions(self):
        """
        If the adding of new assertions works and causes the object to update internal state properly
        """
        test_result = AssertionTestResult()

        test_result.assert_equal(1, 1)
        test_result.assert_equal(2, 2)

        self.assertEqual(2, len(test_result.assertions))
        self.assertEqual(0, test_result.error_count)
        self.assertEqual(0, test_result.exit_code)

    def test_adding_false_assertions(self):
        """
        If adding a false assertion gets registered correctly
        """
        test_result = AssertionTestResult()

        test_result.assert_equal(1, 1)
        test_result.assert_equal(1, 2)

        self.assertEqual(2, len(test_result.assertions))
        self.assertEqual(1, test_result.error_count)
        self.assertNotEqual(0, test_result.exit_code)

    def test_markdown_conversion(self):
        """
        If the conversion to markdown works
        """
        test_result = AssertionTestResult()

        test_result.assert_equal(1, 1)
        test_result.assert_equal(1, 2)
        test_result.assert_equal("hello", "bello")

        markdown_string = test_result.to_markdown()
        print(markdown_string)

        self.assertIsInstance(markdown_string, str)
        self.assertIn('NOT EQUAL', markdown_string)

    def test_pci_read_assertion(self):
        """
        if the custom assertion for the output of a pci_read function works properly
        """
        test_result = AssertionTestResult()

        correct_pci_read = '9010:  000fdf00'
        test_result.assert_pci_read_ok(correct_pci_read)

        error_pci_read = '9010:  000bef00'
        test_result.assert_pci_read_ok(error_pci_read)

        # Now it needs to have one correct assertion and one error!
        self.assertEqual(2, len(test_result.assertions))
        self.assertEqual(1, test_result.error_count)



import tempfile
import os

from unittest import TestCase

from ufotest.util import init_install, get_path
from ufotest.config import CONFIG, reload_config, load_config, Config


class UfotestTestCase(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        # So here I somehow have to setup a new config file etc. I guess I would do this with a temp folder?
        cls.temp_folder = tempfile.TemporaryDirectory()
        cls.folder_path = cls.temp_folder.name

        # Now we need to set a custom path for the installation folder of ufotest, which is inside this temporary
        # folder.
        os.environ['UFOTEST_PATH'] = cls.folder_path
        # After that we can install ufotest into this folder
        init_install()
        # After the installation we also need to reload the config dictionary for the whole system
        cls.config = Config()
        cls.config.reload()

    @classmethod
    def tearDownClass(cls) -> None:
        # We need to properly close the temporary folder here
        cls.temp_folder.cleanup()

    # DEFAULT TESTS
    # -------------

    def test_installation_path(self):
        installation_path = get_path()
        expected_path = self.folder_path
        self.assertEqual(installation_path, expected_path)

    def test_config(self):
        self.assertIn('install', self.config)
        self.assertIn('camera', self.config)



import os
from pathlib import Path

import click

from ufotest.config import CONFIG


PATH = Path(__file__).parent.absolute()
SCRIPTS_PATH = os.path.join(PATH, 'scripts')


SCRIPT_AUTHORS = {
    'michele':              'Michele Caselle <michele.caselle@kit.edu>',
    'jonas':                'Jonas Teufel <jonseb1998@gmail.com',
    'timo':                 'Timo Dritschler <timo.dritschler@kit.edu>'
}


SCRIPTS = {
    'reset': {
        'path':             os.path.join(SCRIPTS_PATH, 'Reset_all.sh'),
        'description':      'Resets the camera parameters to the default state',
        'author':           SCRIPT_AUTHORS['michele']
    },
    'reset_tp': {
        'path':             os.path.join(SCRIPTS_PATH, 'Reset_all_TP.sh'),
        'description':      'Resets the camera parameters to the default state and activates the test pattern',
        'author':           SCRIPT_AUTHORS['michele']
    },
    'status': {
        'path':             os.path.join(SCRIPTS_PATH, 'status.sh'),
        'description':      'Reads out the status parameters of the camera',
        'author':           SCRIPT_AUTHORS['michele']
    },
    'power_up': {
        'path':             os.path.join(SCRIPTS_PATH, 'PWUp.sh'),
        'description':      'Enables the internal power supply of the camera sensor',
        'author':           SCRIPT_AUTHORS['michele']
    },
    'power_down': {
        'path':             os.path.join(SCRIPTS_PATH, 'PWDown.sh'),
        'description':      'Disables the internal power supply of the camera sensor',
        'author':           SCRIPT_AUTHORS['michele']
    },
    'pcie_init': {
        'path':             os.path.join(SCRIPTS_PATH, 'pcie_init.sh'),
        'description':      'Identifies the fpga and initiates the driver for the connection',
        'author':           SCRIPT_AUTHORS['michele']
    },
    'reset_fpga': {
        'path':             os.path.join(SCRIPTS_PATH, 'reset_fpga.sh'),
        'description':      'Resets the fpga',
        'author':           SCRIPT_AUTHORS['michele']
    },
    'reset_dma': {
        'path':             os.path.join(SCRIPTS_PATH, 'dma.sh'),
        'description':      'Resets the dma engine of the fpga',
        'author':           SCRIPT_AUTHORS['michele']
    }
}

# Dynamically registering the scripts from the config file also.
if 'scripts' in CONFIG.keys():
    SCRIPTS.update(CONFIG['scripts'])





"""
A module containing the functionality related to interacting with the camera
"""
import os
import time

import shutil
import click
import numpy as np

from ufotest.config import CONFIG
from ufotest.util import execute_command, get_command_output, execute_script


def pci_write(addr: str, value: str):
    pci_command = 'pci -w {} {}'.format(addr, value)
    exit_code = execute_command(pci_command, verbose=False)
    if exit_code:
        click.secho('Command "{}" failed!'.format(pci_command), fg='red')


def pci_read(addr: str, size):
    pci_command = 'pci -r {} -s {}'.format(addr, str(size))
    value = get_command_output(pci_command)
    return value


def import_raw(path: str, n: int, sensor_width: int, sensor_height: int):
    image = np.fromfile(path, dtype=np.uint16, count=2 * sensor_width * sensor_height * n)
    image = image.reshape((n, sensor_height, sensor_width))
    return image


def set_up_camera(verbose: bool = False):
    # enable the drivers and stuff
    execute_script('pcie_init', verbose=verbose, prefix='sudo ')
    time.sleep(1)
    # Reset all the parameters for the camera
    execute_script('reset_fpga', verbose=verbose)
    time.sleep(1)
    # Enable the sensor power supply
    execute_script('power_up', verbose=verbose)
    time.sleep(1)
    # ?
    execute_script('reset_tp', verbose=verbose)
    time.sleep(1)
    # Display the status just to be save
    execute_script('status', verbose=verbose)

    click.secho('Camera set up finished\n', bold=True)


def tear_down_camera(verbose: bool = False):
    # Disable the sensor power supply
    execute_script('power_down', verbose=verbose)
    # Display the status just to be save
    execute_script('status', verbose=verbose)

    click.secho('camera tear down finished\n', bold=True)


def get_frame(path: str = '/tmp/frame.raw', verbose: bool = False):
    exit_code = save_frame(path, verbose)

    if not exit_code:
        # read the data of the frame and return it
        with open(path, mode='rb+') as file:
            return file.read()
    else:
        return b''


def save_frame(path: str, verbose: bool, tmp_path='/tmp'):
    # Requesting the frame from the camera
    if verbose:
        click.secho('Sending frame request')
    pci_write('0x9040', '0x80000201')
    pci_write('0x9040', '0x80000209')
    # time.sleep(0.1)
    pci_read('9070', '4')
    pci_write('0x9040', '0x80000201')
    time.sleep(0.01)

    # Taking all the raw data of the frame
    if verbose:
        click.secho('Receiving raw data')
    data_path = os.path.join(tmp_path, 'frame.out')
    receive_command = 'pci -r dma0 --multipacket -o {}'.format(data_path)
    execute_command(receive_command, verbose)
    time.sleep(0.1)
    pci_read('9050', '12')
    time.sleep(0.1)

    # Decoding the frame into the RAW format
    if verbose:
        click.secho('Decoding the image, with the following settings:')
        click.secho('-- Camera height:   {}'.format(CONFIG.get_sensor_height()))
        click.secho('-- Camera width:    {}'.format(CONFIG.get_sensor_width()))
    decode_command = 'ipedec -r {height} --num-columns {width} {path} {verbose}'.format(
        height=CONFIG.get_sensor_height(),
        width=CONFIG.get_sensor_width(),
        path=data_path,
        verbose='-v' if verbose else ''
    )
    print(decode_command)
    exit_code = execute_command(decode_command, verbose)
    # Saving the frame to the correct position
    if not exit_code:
        decoded_path = '{}.raw'.format(data_path)
        shutil.move(decoded_path, path)
        if verbose:
            click.secho('Saved frame to "{}"'.format(path), fg='green')
    else:
        click.secho('Error saving frame!', fg='red')

    return exit_code



"""
A module containing various utility functions for usage in other modules.
"""
import string
import click
import random
import re
import subprocess
import importlib.util
from typing import Optional, List
from abc import ABC, abstractmethod

from jinja2 import Template, FileSystemLoader, Environment

from ufotest.config import *
from ufotest.scripts import SCRIPTS

# GLOBAL VARIABLES
VERSION_PATH = os.path.join(PATH, 'VERSION')

TEMPLATE_LOADER = FileSystemLoader(TEMPLATE_PATH)
TEMPLATE_ENVIRONMENT = Environment(loader=TEMPLATE_LOADER)
TEMPLATE_ENVIRONMENT.globals['config'] = Config()


# FUNCTIONS
# =========

def get_repository_name(repository_url: str) -> str:
    """Returns the name of a git repository if the *repository_url* for it is given.

    :param repository_url: the string url of the remote repository

    :return: the string name of the repo, which does NOT include the username it is just the name of the top level
    folder of that repo.
    """
    repository_name = re.findall(r'/([^/]*)\.git', repository_url)[0]
    return repository_name


def markdown_to_html(input_path: str, output_path: str, header_lines: List[str] = []):
    """Converts the markdown file at *input_path* to html file at *output_path*

    The functionality of this function is provided by the command line tool 'md-to-html' provided by the python
    package of the same name.
    This function also offers the additional possibility to add custom string lines to the head section of the created
    html file to potentially insert additional styles.

    :param input_path: The path to the original markdown file
    :param output_path: The path to be created for the html file
    :param header_lines: A list of strings to be included as separate lines into the head of the html
    """
    # The command md-to-html should be pretty self explanatory, it takes a markdown file as the input and produces a
    # corresponding html file from it. We can assume that this command generally works because it is provided by a
    # python module with the same name which is listed as one of the dependencies of this project. So when the ufotest
    # package is installed, this will be installed as well automatically.
    command = 'md-to-html --input {} --output {}'.format(
        input_path,
        output_path
    )
    execute_command(command, False)

    # This plain html file which was created looks pretty plain, so additionally there should be the possibility to
    # add stylesheets to the header of the document. This will be realized by being able to generally insert any kind
    # of string to the header section of the html file. Luckily this will be a rather simple string operation since
    # the previous conversion creates a html file which is very well structured by new lines.
    with open(output_path, mode='r+') as read_file:
        lines = read_file.readlines()

    for line in header_lines:
        lines.insert(lines.index(' </head>'), line)

    with open(output_path, mode='w+') as write_file:
        write_file.write('\n'.join(lines))


def get_version():
    with open(VERSION_PATH) as version_file:
        version = version_file.read()
        version = version.replace(' ', '').replace('\n', '')

    return version


def dynamic_import(module_name: str, file_path: str):
    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module


def get_command_output(command: str, cwd: Optional[str] = None):
    """
    Executes the given "command" and returns the output of the command as string

    The additional "cwd" option can be used to pass a string path description, which is supposed to be used as the
    current working directory for the command execution.

    :param command: The command which is to be executed
    :type command: str

    :param cwd: The path string of the current working directory to be assumed for the command execution
    :type cwd: Optional[str]
    """
    completed_process = subprocess.run(command, cwd=cwd, shell=True, stdout=subprocess.PIPE)
    byte_output = completed_process.stdout
    return byte_output.decode('utf-8')


def execute_command(command: str, verbose: bool, cwd: Optional[str] = None, foreground=True):
    """
    Executes the given system "command"

    The "verbose" flag controls whether or not the output of the command is written to stdout or not. With the "cwd"
    string a path can be passed, which is supposed to be used as the current working directory from which the command
    is to be executed.
    """
    output = None if verbose else subprocess.DEVNULL
    completed_process = subprocess.run(
        command,
        cwd=cwd,
        shell=True,
        stdout=output,
        stderr=output,
        close_fds=not foreground
    )
    return completed_process.returncode


def setup_environment():
    """
    Sets up all the environmental variables defined in the config file
    """
    for key, value in CONFIG['environment'].items():
        # Is this enough?
        os.environ[key] = value


def create_folder(folder_path: str):
    if not os.path.exists(folder_path):
        os.mkdir(folder_path)
        os.chmod(folder_path, 0o777)


def init_install(verbose=False) -> str:
    """Initializes the installation folder for the ufotest app.

    :param verbose: whether or not to print additional info to the console.

    :return: The string absolute path to the top level installation folder which has been created.
    """
    # First we check if the installation folder already exists.
    # "get_path" returns the path to where the config file is supposed to be installed to.
    folder_path = get_path()
    if os.path.exists(folder_path):
        raise IsADirectoryError('An ufotest installation already exists!')
    else:
        # In case it does not exist we will create it
        os.mkdir(folder_path)
        os.chmod(folder_path, 0o777)
        if verbose:
            click.secho('    Created new folder: {}'.format(folder_path))

    # -- CREATE CONFIG FILE
    # Then we need to copy the default template for the config file into this folder, if it is not already there
    config_path = get_config_path()
    if not os.path.exists(config_path):
        shutil.copyfile(CONFIG_TEMPLATE_PATH, config_path)
        if verbose:
            click.secho('    Created default config: {}'.format(folder_path))

    # -- TESTS FOLDER
    # Also we need to create the tests folder inside of this folder
    test_folder_path = os.path.join(folder_path, 'tests')
    os.mkdir(test_folder_path)
    os.chmod(folder_path, 0o777)
    if verbose:
        click.secho('    Created the "tests" folder for ufotest')

    # -- ARCHIVE FOLDER
    # Additionally we need to create a folder for the archive of the test runs
    archive_folder_path = os.path.join(folder_path, 'archive')
    os.mkdir(archive_folder_path)
    os.chmod(folder_path, 0o777)
    if verbose:
        click.secho('    Created the "archive" folder for ufotest')

    # -- BUILDS FOLDER
    # This folder will contain the build reports which are created when a new version of the hardware configuration
    # is built from the source repository
    builds_folder_path = get_builds_path()
    os.mkdir(builds_folder_path)
    if verbose:
        click.secho('    Created the "builds" folder for ufotest')

    # -- STATIC FOLDER
    # This folder contains all the static assets which are required for the internal CI web server. This folder already
    # exists within the ufotest package, it simply needs to be copied into the installation folder...
    static_path = os.path.join(folder_path, 'static')
    shutil.copytree(STATIC_PATH, static_path)
    if verbose:
        click.secho('    Copied the "static" folder into ufotest installation')

    return folder_path


def check_install():
    folder_path = get_path()
    config_path = get_config_path()

    valid_install = True

    if not os.path.exists(folder_path):
        click.secho('ufotest is missing an installation folder!', bold=True, fg='red')
        valid_install = False

    if not os.path.exists(config_path):
        click.secho('ufotest is missing the configuration file at "{}"!'.format(config_path), bold=True, fg='red')
        valid_install = False

    if not valid_install:
        click.secho('A new ufotest installation can be created by typing "ufotest init" into the console...')

    return valid_install


def check_path(path: str, is_dir: bool = False):
    exists = os.path.exists(path)
    correct_type = os.path.isdir(path) if is_dir else os.path.isfile(path)

    return exists and correct_type


def check_vivado():
    vivado_path = CONFIG['install']['vivado_path']
    vivado_settings_path = os.path.join(vivado_path, 'settings64.sh')

    vivado_folder_exists = check_path(vivado_path, is_dir=True)
    vivado_settings_exists = check_path(vivado_settings_path, is_dir=False)

    return vivado_folder_exists and vivado_settings_exists


def execute_script(name: str, prefix: str = '', verbose: bool = False):

    if name not in SCRIPTS.keys():
        click.secho('There is no script with the name "{}" registered!'.format(name), fg='red')
        return

    script = SCRIPTS[name]
    script_folder = os.path.dirname(script['path'])
    script_command = prefix + script['path']

    exit_code = execute_command(script_command, verbose=verbose, cwd=script_folder)
    if not exit_code:
        click.secho('Script "{}" executed successfully!'.format(name), fg='green')
    else:
        click.secho('Script "{}" encountered an error!'.format(name), fg='red')

    return exit_code


def get_template(name: str) -> Template:
    """
    Returns the jinja2 template with the given file name *name* within the static templates folder of this project

    :param name: The file name of the template to be loaded

    :returns: The template object for the specified template
    """
    return TEMPLATE_ENVIRONMENT.get_template(name)


def random_string(length: int):
    letters = string.ascii_letters + ' '
    return ''.join(random.choice(letters) for i in range(length))


def clean_pci_read_output(output: str):
    """
    Cleans the output from the "pci_read" function.

    This essentially only means, that all the newlines and the leading whitespaces are being removed.

    :param output: The output message of the "pci_read" function
    :type output: str

    :return:
    """
    # Removing the newlines
    cleaned = output.replace('\n', '')
    # Removing the whitespaces in the front
    cleaned = cleaned.lstrip(' ')

    return cleaned


# CLASSES
# =======

class AbstractRichOutput(ABC):

    # TO BE IMPLEMENTED
    # -----------------

    @abstractmethod
    def to_markdown(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def to_latex(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def to_string(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def to_html(self) -> str:
        raise NotImplementedError()



from __future__ import annotations
import os
import re
import json
import inspect
import platform
import datetime
import logging
from abc import ABC, abstractmethod
from typing import Tuple, Dict, List, Type, Any, Optional
from contextlib import AbstractContextManager

from ufotest.config import PATH, Config
from ufotest.util import (markdown_to_html,
                          dynamic_import,
                          create_folder,
                          get_template,
                          get_version,
                          AbstractRichOutput)


class TestContext(AbstractContextManager):

    ARCHIVE_FOLDER_FORMAT = 'test_run_%d_%m_%Y__%H_%M_%S'

    def __init__(self, *test_folders):
        # constructed attributes
        self.additional_test_folders = test_folders

        # calculated attributes
        self.config = Config()
        self.archive_folder_path = self.config.get_archive_path()
        self.creation_datetime = datetime.datetime.now()
        self.folder_name = self.creation_datetime.strftime(self.ARCHIVE_FOLDER_FORMAT)
        self.folder_path = os.path.join(self.archive_folder_path, self.folder_name)
        # This general information, about the platform and version on which the test was executed, was previously
        # encapsulated by the TestMetadata class. But now all of this information os combined into the context. This
        # is because the responsibility of the context essentially is to provide such "contextual" information
        self.platform = platform.platform()
        self.version = get_version()  # the version of this software

        static_test_folder = os.path.join(PATH, 'tests')
        dynamic_test_folder = self.config.get_test_folder()
        self.test_folders = [
            static_test_folder,
            dynamic_test_folder,
            *self.additional_test_folders
        ]

        # -- The logging is supposed to be saved into a file within the correct archive folder for this test execution.
        self.logger = logging.Logger('TestContext')
        self.logger.setLevel(logging.DEBUG)
        self.logger_path = os.path.join(self.folder_path, 'context.log')

        # Non initialized attributes
        # So the way this context object works is that a new context object is being created for every testing process
        # the instance of this class will ultimately be passed to the TestRunner object which actually executes the
        # the test process. The responsibility of the context is to provide all the information which is required to
        # specify an individual process. But the responsibility is more than that. The context will also save the
        # results of the test runner. It does that to then later be passed to a TestReport. So for this purpose it has
        # to have some empty properties which are only later being set by the TestRunner.
        self.start_datetime: Optional[datetime.datetime] = None
        self.end_datetime: Optional[datetime.datetime] = None
        self.results = {}
        self.tests = {}
        self.name: Optional[str] = None

    def start(self, name: str):
        self.name = name
        self.start_datetime = datetime.datetime.now()

    def end(self, name: str):
        self.end_datetime = datetime.datetime.now()

    # == AbstractContextManager

    def __enter__(self) -> TestContext:
        # 1 -- CREATING THE ARCHIVE FOLDER
        create_folder(self.folder_path)

        # 2 -- INIT LOGGING
        logger_handler = logging.FileHandler(self.logger_path, mode='w', encoding=None, delay=False)
        self.logger.addHandler(logger_handler)

        # 3 -- LOGGING START MESSAGE
        self.logger.debug('Enter test context')

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # 1 -- LOGGING END MESSAGE
        self.logger.debug('Exit test context')


class TestRunner(object):

    def __init__(self, test_context: TestContext):
        # constructed attributes
        self.context = test_context

        # derived attributes from context
        self.config = self.context.config
        self.logger = self.context.logger

        self.tests = {}
        self.modules = {}

    # == LOADING THE TESTS DYNAMICALLY

    def load_modules(self):
        for test_folder in self.context.test_folders:
            self.load_module_from_test_folder(test_folder)

    def load_module_from_test_folder(self, test_folder: str):
        # All the modules within the test folder have to be dynamically imported to access the TestCase classes within
        # them. This loop iterates all the python modules in the given folder path and then adds the dynamic import to
        # the "modules" dict.
        counter = 0
        for root, dirs, files in os.walk(test_folder):
            for file_name in (file_name for file_name in files if '.py' in file_name):
                file_path = os.path.join(root, file_name)
                module_name = file_name.replace('.py', '')

                module = dynamic_import(module_name, file_path)
                self.modules[module_name] = module
                counter += 1

            # We break here because we dont really want to iterate the whole tree structure. We really only wanted all
            # the modules from the top most folder.
            break

        self.logger.info('loaded {} test modules from folder: {}'.format(counter, test_folder))

    def load_tests(self):
        for module_name, module in self.modules.items():
            self.load_tests_from_module(module)

        self.context.tests = self.tests

    def load_tests_from_module(self, module: Any):
        counter = 1
        # Now we have a dynamically imported module and want to extract the classes from it in a similarly dynamic
        # fashion. Luckily, the "inspect" module provides exactly such functionality.
        # INFO: The second argument to of the "get_members" function takes a function, which essentially acts as a
        # filter for which member exactly to include in the final list.
        for name, cls in inspect.getmembers(module, inspect.isclass):
            # We only want to use those classes, which are subclasses of the AbstractTest base class. This way we are
            # excluding all the possible helper classes which are possibly within those test modules.
            if issubclass(cls, AbstractTest):
                # Now these test modules obviously have to import the abstract base class to make the actual test
                # inherit from them. This is a problem because the previous if statement would also accept these
                # imported duplicates of the base class itself.
                if name != 'AbstractTest':
                    # NOTE: here we are not using the class name as the key of the dict, but the class "property"
                    # "name" which has to be defined by every test case.
                    self.tests[cls.name] = cls
                    counter += 1

        self.logger.info('imported {} tests from the module: {}'.format(counter, module.__name__))

    def load(self):
        self.load_modules()
        self.load_tests()

        self.logger.info('All tests loaded')

    # == TEST SUITES

    def get_test_suite(self, suite_name: str):
        suites = self.config.get_test_suites()
        assert suite_name in suites.keys(), 'Test suite {} unknown!'.format(suite_name)

        test_names = suites[suite_name]
        suite_tests = [test for test_name, test in self.tests.items() if test_name in test_names]
        test_suite = TestSuite(self, suite_tests, suite_name)

        return test_suite

    # == TEST EXECUTION

    def run_test(self, test_name: str):
        assert test_name in self.tests.keys(), 'Test {} does not exist!'.format(test_name)

        self.context.start(test_name)
        # A AbstractTest subclass can be instantiated by passing a single argument to the constructor and that is the
        # the test runner object itself.
        test_class = self.tests[test_name]
        test_instance: AbstractTest = test_class(self)
        test_result: AbstractTestResult = test_instance.execute()

        self.context.results[test_name] = test_result
        self.context.end(test_name)

    def run_suite(self, suite_name: str):
        self.context.start(suite_name)
        test_suite = self.get_test_suite(suite_name)

        results = test_suite.execute_all()
        self.context.results.update(results)
        self.context.end(suite_name)


class _TestRunner(object):
    """
    :deprecated:
    """

    # STATIC LOGGER CONFIG

    def __init__(self, config: Config = Config()):
        self.config = config
        self.modules = {}
        self.tests = {}

        self.archive_folder_path = self.config.get_archive_path()
        # So here I have added a separate step/attribute, which adds the pure folder name string to be saved separately
        # instead of directly assembling it into the absolute path. This value will be needed for example to
        # construct the url to the file server which can be used to view the reports online.
        self.start_time = datetime.datetime.now()
        self.name = self.start_time.strftime('test_run_%d_%m_%Y__%H_%M_%S')
        self.folder_path = os.path.join(
            self.archive_folder_path,
            self.name
        )

        create_folder(self.folder_path)

        # Initializing the logging
        self.logger = logging.Logger('TestRunner')

        # "test_folders" was originally a class attribute and the static and dynamic folder paths where global
        # variables for this folder, but then during testing I realized that CONFIG might not be loaded at that point
        # thus producing a key error for "tests"...
        static_test_folder = os.path.join(PATH, 'tests')
        dynamic_test_folder = self.config['tests']['folder']
        self.test_folders = [
            static_test_folder,
            dynamic_test_folder
        ]

    def load(self):
        for test_folder in self.test_folders:
            # Dynamically import all the modules in this folder and then use inspect to extract all the classes.
            self.load_test_folder_modules(test_folder)

        # Loading the tests from the modules
        for module_name, module in self.modules.items():
            self.load_tests_from_module(module)

    def run_test(self, test_name: str):
        assert test_name in self.tests.keys(), 'Test with the name "{}" does not exists!'.format(test_name)

        test_class = self.tests[test_name]
        test = test_class(self)
        result = test.execute()

        test_report = TestReport({test_name: result}, TestMetadata())
        test_report.save(self.folder_path)
        return test_report

    def run_suite(self, suite_name: str):
        test_suite = self.load_test_suite(suite_name)
        results = test_suite.execute_all()

        test_report = TestReport(results, TestMetadata())
        test_report.save(self.folder_path)
        return test_report

    # HELPER METHODS
    # --------------

    def load_test_folder_modules(self, test_folder: str):
        # First check if this test folder actually exists
        # assert check_path(test_folder, is_dir=True), 'The given test folder path does not refer to an existing folder!'

        # Dynamically import all the modules in this folder and then use inspect to extract all the classes.
        for root, dirs, files in os.walk(test_folder):
            for file in files:
                file_path = os.path.join(root, file)
                module_name = file.replace('.py', '')

                module = dynamic_import(module_name, file_path)
                self.modules[module_name] = module

            break

    def load_tests_from_module(self, module):
        for name, cls in inspect.getmembers(module, inspect.isclass):
            if issubclass(cls, AbstractTest) and name != 'AbstractTest':
                self.tests[cls.name] = cls

    def load_test_suite(self, suite_name: str):
        assert suite_name in self.config['tests']['suites'].keys(), 'The test suite {} is not known!'.format(suite_name)

        test_names = self.config['tests']['suites'][suite_name]
        tests = [self.tests[test_name] for test_name in test_names]

        test_suite = TestSuite(self, tests, suite_name)
        return test_suite


class AbstractTestResult(AbstractRichOutput, ABC):

    def __init__(self, exit_code: int):
        self.exit_code = exit_code
        self.start_datetime = None
        self.end_datetime = None

    def get_status(self):
        if self.passing:
            return 'passing'
        else:
            return 'error'

    @property
    def passing(self) -> bool:
        return self.exit_code == 0

    @property
    def execution_time(self) -> float:
        time_delta: datetime.timedelta = self.end_datetime - self.start_datetime
        return time_delta.total_seconds()


class MessageTestResult(AbstractTestResult):

    def __init__(self, exit_code: int, message: str):
        AbstractTestResult.__init__(self, exit_code)
        self.message = message

    # IMPLEMENT "AbstractRichOutput"
    # ------------------------------

    def to_string(self) -> str:
        return self.message

    def to_markdown(self) -> str:
        return self.message

    def to_latex(self) -> str:
        return self.message

    def to_html(self) -> str:
        return '<div class="message-test-result">{}</div>'.format(self.message)


class AssertionTestResult(AbstractTestResult):

    def __init__(self, detailed: bool = False):
        AbstractTestResult.__init__(self, 0)
        self.detailed = detailed

        self.assertions = []
        self.error_count = 0

    # ASSERTION IMPLEMENTATIONS
    # -------------------------

    def assert_equal(self, expected: Any, actual: Any):
        result = (expected == actual)

        self.assertion_result(
            result,
            '(+) EQUAL "{}" == "{}"'.format(expected, actual),
            '(-) NOT EQUAL "{}" != "{}"'.format(expected, actual)
        )

    def assert_pci_read_ok(self, read_result: str):
        match = re.match('.*:[ ]*000f.*', read_result)
        result = not bool(match)

        self.assertion_result(
            result,
            '(+) PCI READ "{}" IS FINE'.format(read_result),
            '(-) PCI READ "{}" CONTAINS ERROR!'.format(read_result)
        )

    # HELPER METHODS
    # --------------

    def assertion_result(self, result: bool, success_message: str, error_message: str):
        if result:
            self.assertions.append((result, success_message))
        else:
            self.exit_code = 1
            self.error_count += 1
            self.assertions.append((result, error_message))

    # IMPLEMENT "AbstractRichOutput"
    # ------------------------------

    def to_string(self) -> str:
        pass

    def to_markdown(self) -> str:
        template = get_template('assertion_test_result.md')
        return template.render(
            exit_code=self.exit_code,
            assertions=self.assertions,
            detailed=self.detailed,
            error_count=self.error_count
        )

    def to_latex(self) -> str:
        pass

    def to_html(self) -> str:
        return ""


class CombinedTestResult(AbstractTestResult):

    def __init__(self, *test_results):
        self.test_results = test_results
        self.exit_codes = [test_result.exit_code for test_result in self.test_results]
        exit_code = 1 if 1 in self.exit_codes else 0
        AbstractTestResult.__init__(self, exit_code)

    # IMPLEMENT "AbstractRichOutput"
    # ------------------------------

    def to_string(self) -> str:
        return '\n\n'.join(test_result.to_string() for test_result in self.test_results)

    def to_markdown(self) -> str:
        return '\n\n'.join(test_result.to_markdown() for test_result in self.test_results)

    def to_latex(self) -> str:
        return '\\\\ \\\\ \n'.join(test_result.to_latex() for test_result in self.test_results)

    def to_html(self) -> str:
        return ""


class AbstractTest(ABC):

    name = "abstract"
    description = ""

    def __init__(self, test_runner: TestRunner):
        self.test_runner = test_runner
        self.logger = self.test_runner.logger
        self.config = self.test_runner.config

    def execute(self) -> AbstractTestResult:
        start_datetime = datetime.datetime.now()
        try:
            test_result = self.run()
        except Exception as e:
            test_result = MessageTestResult(1, 'Test Execution failed with error "{}"'.format(str(e)))
        end_datetime = datetime.datetime.now()

        test_result.start_datetime = start_datetime
        test_result.end_start_time = end_datetime
        return test_result

    def get_name(self):
        return self.name

    @abstractmethod
    def run(self) -> AbstractTestResult:
        raise NotImplementedError()


class TestSuite(object):

    def __init__(self, test_runner: TestRunner, tests: List[Type[AbstractTest]], name: str):
        self.test_runner = test_runner
        self.tests = tests
        self.suite_name = name

        self.results = {}

    def execute_all(self):
        for test_class in self.tests:
            test = test_class(self.test_runner)
            self.results[test.name] = test.execute()

        return self.results

    def get_name(self):
        return 'suite:{}'.format(self.suite_name)


class TestMetadata(AbstractRichOutput):
    """
    :deprecated:
    """
    def __init__(self, config: Config = Config()):
        self.config = config

        self.date_time = datetime.datetime.now()
        self.platform = platform.platform()
        self.version = get_version()

    def get_date(self):
        date_format = self.config.get_date_format()
        return self.date_time.strftime(date_format)

    def get_time(self):
        time_format = self.config.get_time_format()
        return self.date_time.strftime(time_format)

    # IMPLEMENT "AbstractRichOutput"
    # ------------------------------

    def to_string(self) -> str:
        return ""

    def to_markdown(self) -> str:
        template = get_template('test_meta.md')
        return template.render(
            date=self.get_date(),
            time=self.get_time(),
            platform=self.platform,
            config=self.config,
            version=self.version
        )

    def to_latex(self) -> str:
        return ""

    def to_html(self) -> str:
        return ""


class TestReport(AbstractRichOutput):

    DATETIME_FORMAT = '%d.%m.%Y %H:%M'

    def __init__(self, test_context: TestContext):
        # constructed attributes
        self.context = test_context

        # derived attributes from context
        self.config = self.context.config
        self.platform = str(self.context.platform)
        self.version = self.context.version
        self.results = self.context.results
        self.name = self.context.name
        self.folder_name = self.context.folder_name
        self.folder_path = self.context.folder_path
        self.start = self.context.start_datetime.strftime(self.DATETIME_FORMAT)
        self.end = self.context.end_datetime.strftime(self.DATETIME_FORMAT)
        self.duration = int((self.context.end_datetime - self.context.start_datetime).seconds / 60)
        self.tests = self.context.tests

        # derived attributes from config
        self.repository_url = self.config.get_repository_url()
        self.documentation_url = self.config.get_documentation_url()
        self.sensor = "{} x {} pixels".format(self.config.get_sensor_height(), self.config.get_sensor_width())

        # calculated attributes
        self.test_count = len(self.results)
        self.successful_results = {test_name: result for test_name, result in self.results.items() if result.passing}
        self.successful_count = len(self.successful_results)
        self.error_count = self.test_count - self.successful_count
        self.success_ratio = round(self.successful_count / self.test_count, 2)

    def save(self, folder_path: str):
        # 1 -- SAVE MARKDOWN FILE
        markdown_path = os.path.join(folder_path, 'report.md')
        with open(markdown_path, mode='w') as markdown_file:
            markdown_file.write(self.to_markdown())

        # 2 -- SAVE HTML FILE
        html_path = os.path.join(folder_path, 'report.html')
        with open(html_path, mode='w') as html_file:
            html_file.write(self.to_html())

        # 3 -- SAVE JSON FILE
        json_path = os.path.join(folder_path, 'report.json')
        with open(json_path, mode='w') as json_file:
            json_file.write(self.to_json())

    # == UTILITY FUNCTIONS

    def get_test_description(self, test_name: str) -> str:
        return self.tests[test_name].description

    # == JSON SAVING

    def to_dict(self) -> dict:
        return {
            'platform':             self.platform,
            'version':              self.version,
            'sensor':               self.sensor,
            'name':                 self.name,
            'folder_name':          self.folder_name,
            'folder_path':          self.folder_path,
            'start':                self.start,
            'end':                  self.end,
            'duration':             self.duration,
            'test_count':           self.test_count,
            'successful_count':     self.successful_count,
            'success_ratio':        self.success_ratio
        }

    def to_json(self) -> str:
        data = self.to_dict()
        return json.dumps(data, sort_keys=True, indent=4)

    # == ABSTRACT RICH OUTPUT

    def to_html(self) -> str:
        template = get_template('test_report.html')
        return template.render({'report': self})

    def to_markdown(self) -> str:
        template = get_template('test_report.md')
        return template.render({'report': self})

    def to_string(self) -> str:
        # Not important right now
        pass

    def to_latex(self) -> str:
        # Not important right now
        pass


class _TestReport(AbstractRichOutput):
    """
    :deprecated:
    """
    def __init__(self, test_results: Dict[str, AbstractTestResult], test_meta: TestMetadata):
        self.results = test_results
        self.meta = test_meta

        # COMPUTED ATTRIBUTES
        self.test_count = len(self.results)
        self.passing_count = sum(1 for result in self.results.values() if result.passing)
        self.error_count = sum(1 for result in self.results.values() if not result.passing)
        self.success_ratio = round(self.passing_count / self.test_count, ndigits=5)

    def save(self, path: str):
        """Saves the test report into the folder with the given *path*

        At the current time, the test report will be saved as both a markdown file and an html file which was generated
        from the markdown file. Using a simple web server to serve the archives folder, the test report in the form of
        this html file could then be viewed using a web browser.

        All generated files will have the same base name 'report' but different file endings dependings based on their
        type.

        :param path: the path of the folder into which to save the report
        """
        # Save the report as an markdown file
        markdown_path = os.path.join(path, 'report.md')
        with open(markdown_path, mode='w+') as file:
            file.write(self.to_markdown())

        # Converting the file to a simple html file
        html_path = os.path.join(path, 'report.html')
        markdown_to_html(markdown_path, html_path, [])


    # HELPER METHODS
    # --------------

    def get_file_name(self):
        return

    # IMPLEMENT "AbstractRichOutput"
    # ------------------------------

    def to_string(self) -> str:
        return self.to_markdown()

    def to_markdown(self) -> str:
        template = get_template('test_report.md')
        return template.render(
            report=self,
            meta=self.meta,
            results=self.results
        )

    def to_latex(self) -> str:
        return ""

    def to_html(self) -> str:
        return ""




"""
Module containing the functions to access the configuration of ufotest.
"""
import os
import toml
from pathlib import Path
from typing import List

import shutil
from jinja2 import Environment, PackageLoader, select_autoescape

# The path of the this very python package and the path to the default TOML config file, which will be copied during
# the installation of this project
PATH = Path(__file__).parent.absolute()

TEMPLATE_PATH = os.path.join(PATH, 'templates')
STATIC_PATH = os.path.join(PATH, 'static')
CONFIG_TEMPLATE_PATH = os.path.join(TEMPLATE_PATH, 'default.toml')

# This will be the string path to the HOME folder of the user which is currently executing the script
HOME_PATH = str(Path.home())

# This is the default location of the config folder for this project.
# That would be a new hidden folder ".ufotest" within the home folder of the user.
DEFAULT_PATH = os.path.join(HOME_PATH, '.ufotest')

# This string is the name of the environment variable, which can be used to set a different installation path for the
# '.ufotest' folder.
PATH_ENV = 'UFOTEST_PATH'


def get_builds_path() -> str:
    """Returns the string path of the 'builds' folder within the ufotest installation folder.

    :return: the string absolute path of the builds folder
    """
    path = get_path()
    return os.path.join(path, 'builds')


def get_path() -> str:
    """Returns the path of the installation folder of the ufotest app.

    This installation folder contains for example the following things:
    - The config file for the project. This is a toml file which manages the global config settings for ufotest
    - The 'archive' folder which saves archives of the test reports of all the executed test runs.
    - The 'builds' folder which saves archives of all the reports for the issued builds of the remote source repo.

    :return: the string absolute path to the installation folder
    """
    if PATH_ENV in os.environ.keys():
        return os.environ[PATH_ENV]
    else:
        return DEFAULT_PATH


def get_config_path():
    """
    Returns the path of the config file for ufotest.
    """
    path = get_path()
    return os.path.join(path, 'config.toml')


def load_config():
    """
    Loads the configuration from the config file "config.toml" and returns it as a dictionary
    """
    # First we need to check if a config file even exists.
    # If it does exist, everything is fine and we can load it. If it does not exist at the location
    # were it is supposed to exist we need to create a new one...
    config_path = get_config_path()
    if not os.path.exists(config_path):
        return {}
    else:
        return toml.load(config_path)


def reload_config():
    global CONFIG
    CONFIG = load_config()


class Singleton(type):

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Config(metaclass=Singleton):
    """
    This is a singleton class, which implements the access to the config file.

    **Design Choice**

    So I feel like I want to explain my reasoning for this class here. This is indeed not my first implementation for
    config access within this project. Previously I was simply loading the dictionary into global variable within this
    file and then I could import this global variable and just use the dictionary.

    But while working with that implementation I have found two major problems with it:

    (1) Reloading: The problem is that the dictionary is loaded into the global variable at exactly the point when
    it is imported the first time at some other module. This means that there is no way of globally reloading the
    config to react to file system changes for example. Even if you were to assign the variable with a new dict
    for example, this change would not translate to other files! The reloaded version would only be present in a single
    file. With a singleton of course you can simply solve this problem with a "reload" method.

    (2) Reacting to change: There is a significant downside to using the config dict directly without some sort of
    intermediary layer. If you need some functionality say "CONFIG['camera']['sensor_height']" then this is the path
    which is directly implemented like this in the config file structure. This means that if you were to change the
    structure of the config file you would have to change every occasion in the code, which uses this attribute...
    This is obviously a bad SOC. With a class you could write methods, which wrap certain behaviour. After a change
    only this method would have to be changed.
    """

    def __init__(self):
        # -- LOAD THE DATA FROM FILE
        self.data = load_config()

        # -- INIT CONTEXT
        # So here is the scenario: I have noticed that I need to pass some parameters from the click command call
        # within the terminal all the way down to the email sending sub routine. This is multiple layers and I totally
        # do not want to modify all of them with an additional parameter.
        # This is why I have come up with the 'context' idea. I'll just have a global object where I save the necessary
        # values in and then I can access them in the email subroutine. And the config object is perfect since this
        # already is a global singleton! I'll just add the additional sub dict "context" which will work exactly as
        # I have intended
        self.data['context'] = {
            'hostname':         'localhost',
            'port':             80,
            'verbose':          False
        }

    # IMPLEMENTING DICT FUNCTIONALITY
    # -------------------------------

    def __getitem__(self, item):
        return self.data[item]

    def __setitem__(self, key, value):
        self.data[key] = value

    def __contains__(self, item):
        return item in self.data.keys()

    def keys(self):
        return self.data.keys()

    def values(self):
        return self.data.values()

    def items(self):
        return self.data.items()

    # == WRAPPER METHODS

    def get_repository_url(self):
        return self.data['general']['repository_url']

    def get_documentation_url(self):
        return self.data['general']['documentation_url']

    def get_sensor_width(self):
        sensor_model = self.data['camera']['model']
        return self.data['camera'][sensor_model]['sensor_width']

    def get_sensor_height(self):
        sensor_model = self.data['camera']['model']
        return self.data['camera'][sensor_model]['sensor_height']

    def get_archive_path(self):
        return os.path.expandvars(self.data['tests']['archive'])

    def get_date_format(self):
        return self.data['general']['date_format']

    def get_time_format(self):
        return self.data['general']['time_format']

    def get_test_folder(self):
        return self.data['tests']['folder']

    def get_test_suites(self):
        return self.data['tests']['suites']

    def get_ci_repository_url(self):
        return self.data['ci']['repository_url']

    def get_ci_bitfile_path(self):
        return self.data['ci']['bitfile_path']

    def get_ci_branch(self):
        return self.data['ci']['branch']

    def get_ci_suite(self):
        return self.data['ci']['test_suite']

    def get_email_address(self):
        return self.data['ci']['gmail_address']

    def get_email_password(self):
        return self.data['ci']['gmail_password']

    # == UTILITY METHODS

    def reload(self):
        self.data = load_config()

    def url(self, *paths: str) -> str:
        return '/'.join([
            'http://{}:{}'.format(self.data['context']['hostname'], self.data['context']['port']),
            *paths
        ])

    def static(self, name) -> str:
        return self.url('static', name)


CONFIG = Config()
